#include "MeshObjLoader.h"

#define DEFAULT_SIZE 256

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#define MIN(x, y) ((x) < (y) ? (x) : (y))

static const MeshUV empty_uv(0, 0);
/**
 * Helper Functions
 */
static Polygon* reallocate(Polygon* old, unsigned int oSize, unsigned int nSize)
{
    Polygon* polys = (Polygon*) malloc(sizeof(Polygon) * nSize);
    if (!polys)
    {
	printf("Not enough memory to store mesh.\n");
	assert(0);
    }

    if (old)
    {
	memcpy(polys, old, MIN(oSize, nSize)*sizeof(Polygon));
	free(old);
    }

    return polys;
}

MeshUV::MeshUV(float _u, float _v)
{
    u = _u;
    v = _v;
}


MeshObjLoader::MeshObjLoader()
{
    m_polygons = NULL;
    m_numPolygons = 0;
    m_size = 0;
    m_cleanMemory = true;
}

MeshObjLoader::~MeshObjLoader()
{
    m_vertexList.clear();
    m_uvList.clear();
    m_normalList.clear();
    if (m_cleanMemory) {
	for (unsigned int i = 0; i < m_numPolygons; i++)
	{
	    free(m_polygons[i].vertices);
	    free(m_polygons[i].normals);
	    free(m_polygons[i].uvs);
	}
	free(m_polygons);
    }
}

bool MeshObjLoader::loadMesh(const char* meshfile)
{
    bool valid = true;

    m_size = DEFAULT_SIZE;
    m_polygons = reallocate(m_polygons, m_numPolygons, m_size);
    m_numPolygons = 0;

    m_vertexList.clear();
    m_uvList.clear();
    m_normalList.clear();

    FILE* file = fopen(meshfile, "r");
    if (!file)
    {
	printf("Invalid mesh file %s\n", meshfile);
	return false;
    }

    while (!feof(file))
    {
	/* Parse the line */
	char c[2];
	c[0] = c[1] = '\0';

	fscanf(file, "%2s", c);

	/* Comment */
	if (c[0] == '#')
	    continue;

	/* Vertex */
	if (!strcmp(c, "v"))
	{
	    float x, y, z;
	    if (fscanf(file, "%f %f %f", &x, &y, &z) != 3)
	    {
		printf("Invalid vertex.\n");
		valid = false;
		goto end;
	    }
	    float3 v;
	    v.x = x; v.y = y; v.z = z;
	    m_vertexList.push_back(v);
	}
	/* Texture */
	else if (!strcmp(c, "vt"))
	{
	    float u, v;
	    if (fscanf(file, "%f %f", &u, &v) != 2)
	    {
		printf("Invalid uv.\n");
		valid = false;
		goto end;
	    }
	    m_uvList.push_back(MeshUV(u, v));
	}
	/* Normal */
	else if (!strcmp(c, "vn"))
	{
	    float x, y, z;
	    if (fscanf(file, "%f %f %f", &x, &y, &z) != 3)
	    {
		printf("Invalid normal.\n");
		valid = false;
		goto end;
	    }
	    float3 v;
	    v.x = x; v.y = y; v.z = z;
	    m_normalList.push_back(v);
	}
	/* Face */
	else if (!strcmp(c, "f"))
	{

	    if (m_numPolygons + 1 == m_size)
	    {
		m_size *= 2;
		m_polygons = reallocate(m_polygons, m_numPolygons, m_size);
	    }
	    
	    Polygon* p = &m_polygons[m_numPolygons];
	    m_numPolygons++;


	    int v, t;

	    std::vector<int> faceIndices;
	    std::vector<float3> vertices;
	    std::vector<float3> normals;
	    std::vector<MeshUV> uvs;

	    while (fscanf(file, "%d/%d", &v, &t) == 2) {
		if (v < 1) {
		    valid = false;
		    goto end;
		}
		faceIndices.push_back(v-1);
		vertices.push_back(m_vertexList[v - 1]);
	    }

	    if (vertices.size() < 3)
	    {
		printf("Invalid face.  Can not have polygon with less than 3 vertices.\n");
		valid = false;
		goto end;
	    }

	    /* put it in polygon */
	    unsigned int num = vertices.size();
	    p->vertices = (float*)malloc(sizeof(float3) * num);
	    p->normals = (float*)malloc(sizeof(float3) * num);
	    p->uvs = (MeshUV*) malloc(sizeof(MeshUV) * num);
	    p->faceIndices = (int*) malloc(sizeof(int) * num);

	    p->num_vertices = num;

	    for (unsigned int i = 0; i < num; i++)
	    {
		float3 vertex = vertices[i];

		memcpy(&(p->vertices[i*3]), &vertex, sizeof(float3)); 
		p->faceIndices[i] = faceIndices[i];
	    }
	}
	else
	    continue;
    }

    m_polygons = reallocate(m_polygons, m_numPolygons, m_numPolygons);

end:
    fclose(file);
    return valid;
}

Polygon* MeshObjLoader::getPolygons(unsigned int& numPolygons) 
{
    m_cleanMemory = false;
    numPolygons = m_numPolygons;
    return m_polygons;
}
