package mesh;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.util.Vector;

/**
 * Class representing a simple obj mesh object.
 * Everything will be triangulated on parse.
 * 
 * See main at the bottom to see an example of how this is used.
 * 
 * @author Nong
 */
public class Mesh {

    /**
     * A simple 3-vector to store the position.
     * @author Nong
     */
    public static class Vertex {
	public double x, y, z;
	Vertex() {
	    x = y = z = 0;
	}
	Vertex(double x, double y, double z) {
	    this.x = x;
	    this.y = y;
	    this.z = z;
	}

	public String toString() {
	    return "(" + x + ", " + y + ", " + z + ")";
	}
    }

    /**
     * A 3-tuple to store indices into the vertex list.
     * @author Nong
     */
    public static class Triangle {
	public int A, B, C;
	Triangle(int A, int B, int C) {
	    this.A = A;
	    this.B = B;
	    this.C = C;
	}
    }

    /**
     * Public instance variables for the mesh data.
     */
    public Vector<Vertex>   vertexList;
    public Vector<Triangle> triangleList;

    /**
     * Mesh constructor.     *
     */
    Mesh() {
	vertexList = new Vector<Vertex>();
	triangleList = new Vector<Triangle>();
    }

    /**
     * Parses an obj mesh file.  Returns null if file is invalid.
     * @param file
     * @return
     */    
    static Mesh loadMesh(String file) {
	Mesh mesh = new Mesh();

	try {
	    LineNumberReader r = new LineNumberReader(new FileReader(new File(file)));

	    String line = r.readLine();

	    while (line != null) {

		if (line.length() != 0) {

		    if (line.charAt(0) == 'v') {
			/* Texture */
			if (line.charAt(1) == 't') {
			    // Ignore
			}
			/* Vertex */
			else {
			    String[] temp = line.split(" ");
			    if (temp.length != 4) {
				System.out.println("Invalid file");
				return null;
			    }
			    double x = Double.parseDouble(temp[1]);
			    double y = Double.parseDouble(temp[2]);
			    double z = Double.parseDouble(temp[3]);
			    mesh.vertexList.add(new Vertex(x, y, z));
			}
		    }
		    else if (line.charAt(0) == 'f') {
			String[] temp = line.split(" ");
			if (temp.length == 4) {
			    int A, B, C;
			    String[] temp2;

			    temp2 = temp[1].split("/");
			    A = Integer.parseInt(temp2[0]);

			    temp2 = temp[2].split("/");
			    B = Integer.parseInt(temp2[0]);

			    temp2 = temp[3].split("/");
			    C = Integer.parseInt(temp2[0]);

			    mesh.triangleList.add(new Triangle(A - 1, B - 1, C - 1));
			}
			else if (temp.length == 5) {
			    int A, B, C, D;
			    String[] temp2;

			    temp2 = temp[1].split("/");
			    A = Integer.parseInt(temp2[0]);

			    temp2 = temp[2].split("/");
			    B = Integer.parseInt(temp2[0]);

			    temp2 = temp[3].split("/");
			    C = Integer.parseInt(temp2[0]);

			    temp2 = temp[4].split("/");
			    D = Integer.parseInt(temp2[0]);

			    mesh.triangleList.add(new Triangle(A - 1, B - 1, C - 1));
			    mesh.triangleList.add(new Triangle(C - 1, D - 1, A - 1));
			}
			else {
			    System.out.println("Invalid file");
			    return null;
			}
		    }

		}
		line = r.readLine();
	    }
	}
	catch (IOException e) {
	    System.out.println("Invalid file");
	    return null;
	}

	return mesh;
    }

    /**
     * Some example code to load a mesh and loop through to print the positions of each
     * face.
     * @param args
     */
    public static void main(String[] args) {
	Mesh mesh = Mesh.loadMesh("C:\\data\\CharacterDeformation\\Kim3D\\cage.obj");
	System.out.println("Num Vertices: " + mesh.vertexList.size());
	System.out.println("Num Triangles: " + mesh.triangleList.size());

	/* This code will print out all the vertices of all the triangles for the
	 * mesh
	 */
	/*
	   for (int i = 0; i < mesh.triangleList.size(); i++) {
	   System.out.print(mesh.vertexList.get(mesh.triangleList.get(i).A) + " ");
	   System.out.print(mesh.vertexList.get(mesh.triangleList.get(i).B) + " ");
	   System.out.print(mesh.vertexList.get(mesh.triangleList.get(i).C));
	   System.out.println();
	   }
	   */
    }
}
